open Ast

let gram = Grammar.create (Lex.make ())

let program = Grammar.Entry.create gram "program"
let actions = Grammar.Entry.create gram "actions"

EXTEND
  GLOBAL: program actions;
  program: 
        [ [ decls = LIST1 decl; EOF -> decls ] ]
  ;
  actions:
        [ [ a = LIST0 action -> a ] ]
  ;
  body:
        [ [ "{"; a = LIST0 action; "}" -> a ] ]
  ;
  restricted_attr:
        [ [ n = ID; "="; v = str -> (n, v) ] ]
  ;
  decl:
        [ [ "encoding"; v = str -> 
                Encoding v 
	  | "namespace"; n = ID; "="; v = str ->
                Namespace (n, v)
          | "template"; m = str; a = attrs; acs = body -> 
                Template (("match", m) :: a, acs)
          | "output"; a = LIST0 restricted_attr -> Output a
          | var_kw; n = any_name; "="; v = single_action -> Variable (n, v)
          | "def"; n = ID; "("; args = LIST0 xattr SEP ","; ")"; 
            a = attrs; bd = body -> 
                DefTemplate (("name", n) :: a, args, bd)
          | "mode"; m = str; "{"; decls = LIST0 decl; "}" ->
                Mode (m, decls)
          | "macro"; n = ID; "("; args = LIST0 xattr SEP ","; ")";
            bd = body ->
                Macro (n, args, Container bd)
        ] ]
  ;
  var_kw: [ [ "var" | "variable" ] ] ;
  str:
        [ [ s = Q -> s ] ]
  ;
  single_action:
        [ [ s = Q -> String s
          | s = E -> Entity s
          | s = A -> MacroArg s
          | v = V -> ValueOf v
          | t = T -> QString t
          | a = body -> Container a ] ]
  ;
  attr:
        [ [ n = any_name; "="; v = str -> (n, v) ] ]
  ;
  xattr:
        [ [ n = any_name; "="; v = single_action -> (n, v) ] ]
  ;
  attrs: 
        [ [ x = LIST0 attr -> x ] ] 
  ;
  else_clause:
        [ [ "else"; b = body -> Some b | (* nothing *) -> None ] ] 
  ;
  when_clause:
        [ [ "when"; test = str; a = attrs; bd = body ->
                When (("test", test) :: a, Container bd)
          | "otherwise"; a = attrs; bd = body ->
                Otherwise (a, Container bd)
          | "else"; a = attrs; bd = body ->
                Otherwise (a, Container bd)
        ] ]
  ;
  action:
        [ [ a = single_action -> a
          | "%"; name = any_name; at = attrs; bd = body ->
                Node (name, at, bd)
          | "<"; name = any_name; at = attrs; ">" ->
                NodeStart (name, at)
          | "<"; name = any_name; at = attrs; "/"; ">" ->
                Node (name, at, [])
          | "<"; "/"; name = any_name; ">" ->
                NodeEnd name
          | v = V ->
                ValueOf v
          | "nbsp" ->
                Entity "nbsp"
          | "apply"; a = attrs; empty_tag ->
                Apply a
          | n = ID; "("; at = LIST0 xattr SEP ","; ")" ->
                CallTemplate (n, at)
          | "%"; name = ID; "("; at = LIST0 xattr SEP ","; ")" ->
                CallMacro (name, at)
          | "macro-arg"; name = ID ->
                MacroArg name
          | "redirect"; file = str; a = attrs; bd = body ->
                Redirect ((("select", file) :: a), Container bd)
          | "if"; cond = str; a = LIST0 attr; bd = body; el = else_clause ->
            begin
              match el with
              | None -> 
                If ((("test", cond) :: a), Container bd)
              | Some e -> 
                IfElse ((("test", cond) :: a), Container bd, Container e)
            end
          | "for-each"; sel = str; a = attrs; bd = body ->
            ForEach (("select", sel) :: a, Container bd)
          | "choose"; a = attrs; "{"; w = LIST0 when_clause; "}" ->
            Choose (a, w)
          | var_kw; n = any_name; "="; v = single_action -> AVariable (n, v)
        ] ]
  ;
  empty_tag:
        [ [ "{"; "}" | ";" ] ]
  ;
  any_name:
        [ [ a = ID -> a
          | "apply" -> "apply"
          | "choose" -> "choose"
          | "def" -> "def"
          | "else" -> "else"
          | "encoding" -> "encoding"
          | "for-each" -> "for-each"
          | "if" -> "if"
          | "macro" -> "macro"
          | "macro-arg" -> "macro-arg"
          | "mode" -> "mode"
          | "namespace" -> "namespace"
          | "nbsp" -> "nbsp"
          | "otherwise" -> "otherwise"
          | "output" -> "output"
          | "redirect" -> "redirect"
          | "template" -> "template"
          | "var" -> "var"
          | "variable" -> "variable"
          | "when" -> "when"
        ] ]
  ;
END

let err_msg (x, y, err, file, str) =
  let b = Buffer.create 80 in
  let s = Stream.of_string str in
  let rec find_line ln x =
    let c = Stream.next s in
    if c = '\n' then
      if x < 0 then
        (-x, ln, Buffer.contents b)
      else
        begin
          Buffer.clear b;
          find_line (ln + 1) (x - 1)
        end
    else
      begin
        Buffer.add_char b c;
        find_line ln (x - 1)
      end in
  let (ch, line_no, line) = 
    try
      find_line 1 x
    with Stream.Failure -> (0, 0, "") in
  prerr_endline (file ^ ":" ^ (string_of_int line_no) ^ ": " ^ err);
  if line <> "" then
    begin
      prerr_endline (">>>  " ^ line);
      let pref = String.length line - ch - 1 in
      prerr_endline (">>>  " ^ (String.make pref ' ') ^ 
                               (String.make (y - x + 1) '^'))
    end
  else
    ();
  exit 1

let parse_actions str =
  try
    Grammar.Entry.parse actions (Stream.of_string str)
  with 
    | Stdpp.Exc_located ((x, y), Stream.Error err) 
    | Stdpp.Exc_located ((x, y), Token.Error err) ->
        err_msg (x, y, err, "qq", str)

let parse file str =
  try
    Grammar.Entry.parse program (Stream.of_string str)
  with 
    | Stdpp.Exc_located ((x, y), Stream.Error err) 
    | Stdpp.Exc_located ((x, y), Token.Error err) ->
        err_msg (x, y, err, file, str)
