type ctx = {
  ctx_buffer: Buffer.t;
  ctx_keywords: (string, unit) Hashtbl.t;
  mutable ctx_pos: int;
}

let initial_ctx () = 
  {ctx_pos = 0; 
   ctx_buffer = Buffer.create 80;
   ctx_keywords = Hashtbl.create 16}

let rec lexit ctx s =
  let is_id_beg c =
    match c with
    | 'a'..'z' | 'A'..'Z' | '_' | '-' | ':' -> true
    | _ -> false in

  let is_digit c =
    match c with
    | '0'..'9' -> true
    | _ -> false in

  let getc ctx s =
    let c = Stream.next s in
    let _ = ctx.ctx_pos <- ctx.ctx_pos + 1 in 
    c in
  
  let end_delim c =
    match c with
    | '{' -> '}'
    | '(' -> ')'
    | '[' -> ']'
    | '<' -> '>'
    | _ when (String.contains "!@#$%^&*-+=;:\"',./?|" c) -> c
    | _ -> raise Not_found in
    
  let check_q s =
    match Stream.peek s with
    | Some c ->
      begin
        try 
          let _ = end_delim c in 
          true 
        with Not_found -> false
      end
    | _ -> false in
    
  let rec ident ctx s =
    match Stream.peek s with
    | Some c when ((is_id_beg c) || (is_digit c)) ->
      let _ = Buffer.add_char ctx.ctx_buffer (getc ctx s) in
      ident ctx s
    | _ -> () in
      
  let rec eat_till stopper ctx s =
    match Stream.peek s with
    | Some c when (c <> stopper) ->
      let _ = Buffer.add_char ctx.ctx_buffer (getc ctx s) in
      eat_till stopper ctx s
    | Some c when (c = stopper) ->
      let _ = getc ctx s in ()
    | _ -> () in
     
  let sublex ch fnc cons  =
    Buffer.clear ctx.ctx_buffer;
    begin
      match ch with
      | Some c -> Buffer.add_char ctx.ctx_buffer c
      | _ -> ();
    end;
    let start_ch = ctx.ctx_pos in
    let _ = fnc ctx s in
    let stop_ch = ctx.ctx_pos in
    ((cons, Buffer.contents ctx.ctx_buffer), (start_ch, stop_ch)) in

  try
    let c = getc ctx s in
    match c with
    | ' ' | '\t' | '\n' | '\r' -> lexit ctx s
    | '#' ->
      let _ = eat_till '\n' ctx s in
      lexit ctx s
    | 'q' when (check_q s) ->
      sublex None (eat_till (end_delim (getc ctx s))) "Q"
    | 'v' when (check_q s) ->
      sublex None (eat_till (end_delim (getc ctx s))) "V"
    | 'e' when (check_q s) ->
      sublex None (eat_till (end_delim (getc ctx s))) "E"
    | 'a' when (check_q s) ->
      sublex None (eat_till (end_delim (getc ctx s))) "A"
    | 't' when (check_q s) ->
      sublex None (eat_till (end_delim (getc ctx s))) "T"
    | '\'' ->
      sublex None (eat_till '\'') "Q"
    | '"' ->
      sublex None (eat_till '"') "Q"
    | '=' | '%' | ':' | '{' | '}' | '(' | ')' | ',' | '/' | '<' | '>' 
    | ';'
    ->
      (("", String.make 1 c), (ctx.ctx_pos - 1, ctx.ctx_pos))
    | _ -> 
      if is_id_beg c then
        let ((_, s), loc) = sublex (Some c) ident "" in
        let cons =
          try
            Hashtbl.find ctx.ctx_keywords s;
            ""
          with Not_found -> "ID" in
        ((cons, s), loc)
      else
        raise (Token.Error "illegal character")
   with Stream.Failure ->
    (("EOF", ""), (ctx.ctx_pos, ctx.ctx_pos))
        

let rec printer lex s =
  let ((a, b), (c, d)) = lex s in
  let _ = Printf.printf "(%s, %s), (%d, %d)\n" a b c d in
  match a with
  | "EOF" -> ()
  | _ -> printer lex s

let tparse _ = None

let using ctx (con, txt) =
  match con with
  | "" ->
    begin
      try
        let _ = Hashtbl.find ctx.ctx_keywords txt in
        ()
      with Not_found ->
        Hashtbl.add ctx.ctx_keywords txt ()
    end
  | "T" | "A" | "V" | "Q" | "E" | "ID" | "EOF" -> ()
  | _ -> raise (Token.Error ("bad constructor \"" ^ con ^ "\""))

let removing ctx (con, txt) =
  if con = "" then Hashtbl.remove ctx.ctx_keywords txt else ()

let text (con, txt) =
  if con = "" then txt
  else con ^ "(" ^ txt ^ ")"
  
let make () =
  let ctx = initial_ctx () in
  {Token.func = Token.lexer_func_of_parser (lexit ctx);
   Token.using = using ctx;
   Token.removing = removing ctx;
   Token.tparse = tparse;
   Token.text = text}

;;
