open Ast

exception Error of string

type xml_node =
  | XNode of string * (attr list) * (xml_node list)
  | XText of string
  | XContainer of xml_node list

type ctx = {
  ctx_encoding: string;
  ctx_output: xml_node list;
  ctx_mode: string;
  ctx_macros: (string, decl) Hashtbl.t;
  ctx_macro_args: (string * action) list;
  ctx_uses_redirect: bool;
  ctx_namespaces: (string, string) Hashtbl.t
}

let init_ctx () = {
  ctx_encoding = ""; 
  ctx_output = [];
  ctx_mode = "";
  ctx_macros = Hashtbl.create 16;
  ctx_macro_args = [];
  ctx_uses_redirect = false;
  ctx_namespaces = Hashtbl.create 1
}

let list_apply ctx f elems =
  let rec tail_apply res ctx elems  =
    match elems with
    | x::xs ->
      let (ctx, fx) = f ctx x in
      tail_apply (fx::res) ctx xs
    | [] -> (ctx, List.rev res) in
  tail_apply [] ctx elems

let amp_rx = Str.regexp "&"
let lt_rx = Str.regexp "<" 
let gt_rx = Str.regexp ">"
let quot_rx = Str.regexp "\""

let xml_escape str =
  let str = Str.global_replace amp_rx "&amp;" str in
  let str = Str.global_replace lt_rx "&lt;" str in
  let str = Str.global_replace gt_rx "&gt;" str in
  let str = Str.global_replace quot_rx "&quot;" str in
  str

let string_of_attrs attrs =
  String.concat " "
    (List.map (fun (n, v) -> n ^ "=\"" ^ (xml_escape v) ^ "\"") attrs)

let fixup_mode ctx attrs =
  if (ctx.ctx_mode = "") ||
     (List.exists (fun (x, _) -> x = "mode") attrs) then
       attrs
  else
       ("mode", ctx.ctx_mode) :: attrs
       
let rec compile_decl ctx decl =
  let (ctx, n) =
    match decl with
    | Encoding e -> 
        ({ctx with ctx_encoding = e}, [])
    | Namespace (n, v) ->
        let _ = Hashtbl.add ctx.ctx_namespaces n v in
	(ctx,[])
    | Template (attrs, actions) ->
        let (ctx, children) = list_apply ctx compile_action actions in
        let attrs = fixup_mode ctx attrs in
        let n = XNode ("xsl:template", attrs, children) in
        (ctx, [n])
    | DefTemplate (attrs, args, actions) ->
        let (ctx, children) = list_apply ctx compile_action actions in
        let attrs = fixup_mode ctx attrs in
        let args = parms ctx "xsl:param" [] args in
        let n = XNode ("xsl:template", attrs, args @ children) in
        (ctx, [n])
    | Output a ->
        let n = XNode ("xsl:output", a, []) in
        (ctx, [n])
    | Variable (name, value) ->
        let (ctx, node) = compile_action ctx value in
        let res = XNode ("xsl:variable", [("name", name)], [node]) in
        (ctx, [res]) 
    | Mode (mode, decls) ->
        let old_mode = ctx.ctx_mode in
        let ctx = {ctx with ctx_mode = mode} in
        let ctx = List.fold_left compile_decl ctx decls in
        ({ctx with ctx_mode = old_mode}, []) 
    | Macro (name, args, actions) ->
        let _ = Hashtbl.add ctx.ctx_macros name decl in
        (ctx, [])
  in
  {ctx with ctx_output = n @ ctx.ctx_output}

and remove_newlines s = 
  try
    while true do
      String.set s (String.index s '\n') ' '
    done;
    assert false
  with Not_found -> s
  
and remove_tabs s = 
  try
    while true do
      String.set s (String.index s '\t') ' '
    done;
    assert false
  with Not_found -> s
  
and compile_action ctx act =
  match act with
  | String s -> 
    (ctx, XText s)
  | Node (n, a, b) ->
    let (ctx, b') = list_apply ctx compile_action b in
    (ctx, XNode (n, a, b'))
  | NodeStart (n, a) -> 
    (ctx, XText ("<" ^ n ^ " " ^ (string_of_attrs a) ^ ">"))
  | NodeEnd n -> 
    (ctx, XText ("</" ^ n ^ ">"))
  | ValueOf s ->
    (ctx, XNode ("xsl:value-of", 
                 [("select", remove_tabs (remove_newlines s))], 
                 []))
  | Entity s ->
    (ctx, XNode ("xsl:text", 
                 [("disable-output-escaping","yes")], 
                 [XText ("&amp;" ^ s ^ ";")]))
  | QString s ->
    (ctx, XNode ("xsl:text", 
                 [("disable-output-escaping","yes")], 
                 [XText (xml_escape s)]))
  | CallTemplate (n, attrs) ->
    (ctx, XNode ("xsl:call-template", 
                 [("name", n)], 
                 parms ctx "xsl:with-param" [] attrs))
  | Container a ->
    let (ctx, b') = list_apply ctx compile_action a in
    (ctx, XContainer b')
  | Eval s ->
    let (ctx, b') = list_apply ctx compile_action (Gram.parse_actions s) in
    (ctx, XContainer b')
  | Apply attrs ->
    let attrs = fixup_mode ctx attrs in
    let attrs = List.filter (fun (x, y) -> x <> "mode" || y <> "") attrs in
    (ctx, XNode ("xsl:apply-templates", attrs, []))
  | CallMacro (n, args) ->
    let d = 
      try 
        Hashtbl.find ctx.ctx_macros n
      with Not_found -> raise (Error ("call to undefined macro: " ^ n)) in
    let (formal, body) =
      match d with
      | Macro (_, a, d) -> (a, d)
      | _ -> assert false in
    let old_macro_args = ctx.ctx_macro_args in
    let ctx = {ctx with ctx_macro_args = args @ formal} in
    let (ctx, node) = compile_action ctx body in
    ({ctx with ctx_macro_args = old_macro_args}, node)
  | MacroArg n ->
    let v =
      try
        List.assoc n ctx.ctx_macro_args
      with Not_found -> 
        raise (Error ("reference to undefined macro variable: " ^ n)) in
    compile_action ctx v
  | Redirect (attrs, body) ->
    let (ctx, n) = compile_action ctx body in
    ({ctx with ctx_uses_redirect = true}, 
     XNode ("redirect:write", attrs, [n]))
  | If (attrs, body) ->
    let (ctx, n) = compile_action ctx body in
    (ctx, XNode ("xsl:if", attrs, [n]))
  | IfElse (attrs, then_body, else_body) ->
    let (ctx, then_node) = compile_action ctx then_body in
    let (ctx, else_node) = compile_action ctx else_body in
    (ctx, XNode ("xsl:choose", [], 
                   [XNode ("xsl:when", attrs, [then_node]);
                    XNode ("xsl:otherwise", [], [else_node])]))
  | ForEach (attrs, body) ->
    let (ctx, n) = compile_action ctx body in
    (ctx, XNode ("xsl:for-each", attrs, [n]))
  | Choose (attrs, w) ->
    let (ctx, b') = list_apply ctx compile_when_clause w in
    (ctx, XNode ("xsl:choose", attrs, b'))
  | AVariable (name, value) ->
    let (ctx, node) = compile_action ctx value in
    (ctx, XNode ("xsl:variable", [("name", name)], [node]))

and compile_when_clause ctx w =
  match w with
  | When (attrs, body) ->
    let (ctx, n) = compile_action ctx body in
    (ctx, XNode ("xsl:when", attrs, [n]))
  | Otherwise (attrs, body) ->
    let (ctx, n) = compile_action ctx body in
    (ctx, XNode ("xsl:otherwise", attrs, [n]))
  
and parms ctx node res lst =
  match lst with
  | (n, v) :: xs ->
    let (ctx, compiled) = compile_action ctx v in
    let n = XNode (node, [("name", n)], [compiled]) in
    parms ctx node (n :: res) xs
  | [] -> List.rev res 

let rec output_xnode ch n =
  let p = output_string ch in
  match n with
  | XText s -> p s
  | XContainer x ->
    List.iter (output_xnode ch) x
  | XNode (n, attrs, children) ->
    p ("<" ^ n ^ " ");
    p (string_of_attrs attrs);
    match children with
    | [] -> 
       p "/>"
    | _ -> 
       p ">";
       List.iter (output_xnode ch) children;
       p ("</" ^ n ^ ">\n")
  
let compile decls outch =
  let ctx = List.fold_left compile_decl (init_ctx ()) decls in
  let xmlns = 
    List.rev (Hashtbl.fold  
                 (fun n v l -> ("xmlns:" ^ n, v) :: l)
                 ctx.ctx_namespaces [])
  in
  let redir_stuff =
    if ctx.ctx_uses_redirect then
      [("xmlns:redirect", "org.apache.xalan.xslt.extensions.Redirect");
       ("extension-element-prefixes", "redirect")]
    else [] in
  let n = XNode ("xsl:stylesheet", 
                 [("version", "1.0");
                  ("xmlns:xsl", "http://www.w3.org/1999/XSL/Transform")] @
		  xmlns @
                 redir_stuff,
                 List.rev ctx.ctx_output) in
  let hd = "<?xml version=\"1.0\" " ^ 
                (if ctx.ctx_encoding = "" then 
                   "" 
                 else
                   "encoding=\"" ^ ctx.ctx_encoding ^ "\" ") ^ "?>\n" in
  let _ = output_string outch hd in
  let _ = output_xnode outch n in
  ()
